/******************************************************************************
 * $Id: handlers.c,v 1.45 2001/01/01 23:48:31 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Event/signal handlers.
 ******************************************************************************/

#include "header.h"

#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

/*****************************************************************************
 * Handler for the key press event (in main widget).
 *
 * Parms   : pWidget
 *           Ie pWidgetMain.
 *****************************************************************************/
gboolean
HandlerKeyPressMain( GtkWidget*   pWidget,
                     GdkEventKey* pEvent,
                     gpointer     unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return TRUE;

  /*
   * F3 /Find/Text ?
   * or Ctrl+F
   */
#if 0
   if ( pEvent->keyval == GDK_F3
#else
   if (     pEvent->keyval == GDK_F3
        || (pEvent->keyval == GDK_f  &&  pEvent->state & GDK_CONTROL_MASK) )
#endif
   {
     /*
      * Otherwise, Ctrl+F move cursor to right in text widget,
      * which will disrupt search which depends on cursor position.
      */
      pEvent->keyval = KLUDGE_IGNORED_GDK_KEY;  /* <-- KLUDGE */

      CreateFindDialog( MENU_ITEM_FIND_TEXT );
      return TRUE;
   }

  /*
   * F4 /Find/Tag ?
   */
   if ( pEvent->keyval == GDK_F4 )
   {
      CreateFindDialog( MENU_ITEM_FIND_TAG );
      return TRUE;
   }

  /*
   * F5 /Goto/Tag ?
   * or Ctrl+Enter
   */
   if (     pEvent->keyval == GDK_F5
        || (pEvent->keyval == GDK_Return  &&  pEvent->state & GDK_CONTROL_MASK) )
   {
      SetSensitivityWhileBusy( FALSE );
      GotoTagUnderCursor();
      SetSensitivityWhileBusy( TRUE );
      return TRUE;
   }

  /*
   * F6 Earlier Tag ?
   */
   if ( pEvent->keyval == GDK_F6 )
   {
      HandlerToolbarButtonEarlierTag( NULL, NULL );
      return TRUE;
   }

  /*
   * F7 Later Tag ?
   */
   if ( pEvent->keyval == GDK_F7 )
   {
      HandlerToolbarButtonLaterTag( NULL, NULL );
      return TRUE;
   }

  /*
   * Ctrl+N Find Next (text or tag)
   * This case also appears in key handler of text widget (below)
   * because Ctrl+N is CursorDown in context of text editing.
   */
   if ( (pEvent->keyval == GDK_n && pEvent->state & GDK_CONTROL_MASK) )
   {
     /*
      * The KLUDGE at text widget's key handler text widget must be
      * repeated here.
      */
      gboolean findReverseCopy = findReverse;
      findReverse = FALSE;

      pEvent->keyval = KLUDGE_IGNORED_GDK_KEY;  /* <-- KLUDGE */
      FindNextTextOrTag();

      findReverse = findReverseCopy;

      return TRUE;
   }

  /*
   * Ctrl+R Reverse Find Next (text or tag)
   */
   if ( pEvent->keyval == GDK_r )
   {
      gboolean findReverseCopy = findReverse;
      findReverse = TRUE;

      FindNextTextOrTag();

      findReverse = findReverseCopy;
      return TRUE;
   }

  /*
   * Return FALSE to defer event to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for both key press/release events in text widget.
 *
 * Parms   : pWidget
 *           Ie pWidgetText.
 *****************************************************************************/
gboolean
HandlerKeyPressReleaseText( GtkWidget*   pWidget,
                            GdkEventKey* pEvent,
                            gpointer     unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return TRUE;

  /*
   * Ctrl+N Find Next (text or tag)
   */
   if ( (pEvent->keyval == GDK_n && pEvent->state & GDK_CONTROL_MASK) )
   {
     /*
      * KLUDGE: Even though TRUE is returned to stop propagating this event,
      *         Ctrl+N will be handled by another handler as Cursor Down.
      *         Kludge is to modify event struct to pretend as though
      *         an unhandled key was pressed.
      */
      gboolean findReverseCopy = findReverse;
      findReverse = FALSE;

      pEvent->keyval = KLUDGE_IGNORED_GDK_KEY;  /* <-- KLUDGE */
      FindNextTextOrTag();

      findReverse = findReverseCopy;

      return TRUE;
   }

  /*
   * Ctrl+L (emulate emacs)
   */
   if ( (pEvent->keyval == GDK_l && pEvent->state & GDK_CONTROL_MASK) )
   {
      RecenterTextWidget();
      return TRUE;
   }

  /*
   * Complain if user is trying to edit a file (which can't be saved).
   */
   if (    (pEvent->keyval != GDK_Alt_L)
        && (pEvent->keyval != GDK_Alt_R)
        && (pEvent->keyval != GDK_Shift_L)
        && (pEvent->keyval != GDK_Shift_R)
        && (pEvent->keyval != GDK_Control_L)
        && (pEvent->keyval != GDK_Control_R)
        && (pEvent->keyval != GDK_Up)
        && (pEvent->keyval != GDK_Down)
        && (pEvent->keyval != GDK_Left)
        && (pEvent->keyval != GDK_Right)
        && (pEvent->keyval != GDK_Page_Up)
        && (pEvent->keyval != GDK_Page_Down)
        && (pEvent->keyval != GDK_Home)
        && (pEvent->keyval != GDK_End)
        && (pEvent->keyval != GDK_Escape)
        && (pEvent->keyval != GDK_F1)
        && (pEvent->keyval != GDK_F2)
        && (pEvent->keyval != GDK_F3)
        && (pEvent->keyval != GDK_F4)
        && (pEvent->keyval != GDK_F5)
        && (pEvent->keyval != GDK_F6)
        && (pEvent->keyval != GDK_F7)
        && (pEvent->keyval != GDK_F8)
        && (pEvent->keyval != GDK_F9)
        && (pEvent->keyval != GDK_F10)
        && (pEvent->keyval != GDK_F11)
        && (pEvent->keyval != GDK_F12)
        && ! (pEvent->state & GDK_CONTROL_MASK)
        && ! (pEvent->state & GDK_MOD1_MASK) )
   {
      static gboolean shownBefore = FALSE;

      if ( ! shownBefore )
      {
         Popup( "Editing files is not supported!\nFiles can't be saved.",
                "Warning",
                FONT_PROPORTIONAL );
         shownBefore = TRUE;
      }
      /* fall thru */
   }

  /*
   * Return FALSE to defer event to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for the key press event (in a Find dialog widget).
 *
 * Parms   : pWidget
 *           Ie pWidgetDialogFind.
 *****************************************************************************/
gboolean
HandlerKeyPressDialogFind( GtkWidget*   pWidget,
                           GdkEventKey* pEvent,
                           gpointer     unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return TRUE;

  /*
   * ESC ?
   */
   if ( pEvent->keyval == GDK_Escape )
   {
     /*
      * If a dialog is open then close it.
      */
      if ( pWidgetDialogFind )
         gtk_widget_destroy( pWidgetDialogFind );
      return TRUE;
   }

  /*
   * Enter ?
   */
   if ( pEvent->keyval == GDK_Return )
   {
      FindDialogClickedFind();
      return TRUE;
   }

  /*
   * Alt+F ? (select "_Find" button in Find dialog)
   */
   if ( (pEvent->keyval == GDK_f && pEvent->state & GDK_MOD1_MASK) )
   {
      FindDialogClickedFind();
      return TRUE;
   }

  /*
   * Alt+W ? (select "_Word at Cursor" button in Find dialog)
   * Ignore if button wasn't created.
   */
   if (    (pEvent->keyval == GDK_w && pEvent->state & GDK_MOD1_MASK)
        && buttonWordAtCursorFlag )
   {
      WordAtCursorClicked();
      return TRUE;
   }

  /*
   * Alt+P ? (select "Restore Cursor _Position" button in Find dialog)
   * Ignore if button wasn't created.
   */
   if (    (pEvent->keyval == GDK_p && pEvent->state & GDK_MOD1_MASK)
        && buttonRestoreCursorFlag )
   {
      RestoreCursorClicked();
      return TRUE;
   }


  /*
   * Return FALSE to defer event to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for the key press event in line # dialog widget.
 *
 * Parms   : pWidget
 *           Ie pWidgetDialogLine.
 *****************************************************************************/
gboolean
HandlerKeyPressDialogLine( GtkWidget*   pWidget,
                           GdkEventKey* pEvent,
                           gpointer     unused )
{
  /*
   * ESC ?
   */
   if ( pEvent->keyval == GDK_Escape )
   {
     /*
      * Close dialog.
      */
      gtk_widget_destroy( pWidget );
      return TRUE;
   }

  /*
   * Return FALSE to defer event to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for when the goto line combo widget is activated (ie Enter pressed).
 *
 * Parms   : pWidget
 *           Pointer to text entry widget of /Goto/Line dialog.
 *****************************************************************************/
gboolean
HandlerActivateComboLine( GtkWidget*   pWidget,
                          GdkEventKey* pEvent,
                          gpointer     unused )
{
   int	 line;
   char* pEntryText;

  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return TRUE;

  /*
   * The line # the user wishes to goto is in the entry widget.
   * Convert the text string to a binary number.
   */
   pEntryText = gtk_editable_get_chars( GTK_EDITABLE(GTK_ENTRY(GTK_COMBO(pWidgetComboLine)->entry)),
                                        0, -1 );

   line = atoi( (const char*)pEntryText ) - 1;

  /*
   * Ensure the line # is valid.
   */
   if ( line >= 0  &&  line < MAX_LINES )
   {
     /*
      * Scroll text widget to requested line #.
      */
      if ( HighlightLine( line ) )
         ScrollTextWidgetToLine( line, SCROLL_TEXT_VERT_FRAC );
      else
         PrintStatusbar( "line # is invalid" );

     /*
      * Record the string the user typed.
      */
      PrependComboDropDown( pEntryText,
                            &pGlistLineStrings );
   }
   else
      PrintStatusbar( "Requested line # is invalid." );

  /*
   * Close dialog.
   */
   gtk_widget_destroy( pWidgetDialogLine );

   g_free( pEntryText );

  /*
   * Event was handled here, return TRUE.
   */
   return TRUE;
}

/*****************************************************************************
 * Handler for updating the line #.
 *
 * Notes: The reason that an idle function prints the line # is to allow
 *        GTK+ to finish updating the text widget, esp. when cursor keys
 *        are pressed.  This is why the key event handler LineNumHandler()
 *        enables and defers to this idle function.
 *****************************************************************************/
gboolean
HandlerLineNum( GtkWidget*      pWidget,
                GdkEventButton* pEvent,
                gpointer        ununsed )
{
  /*
   * Enable an idle function to be called after GTK+ becomes idle,
   * which is the correct time to accurately update the line #.
   */
   gtk_idle_add( IdlePrintLineNum,
                 NULL );

  /*
   * Return FALSE to defer event to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for the button "Earlier Tag".
 *****************************************************************************/
void
HandlerToolbarButtonEarlierTag( GtkWidget* pWidgetUnused,
                                gpointer   unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return;

   SetSensitivityWhileBusy( FALSE );
   StepHistoryList( 1 );
   SetSensitivityWhileBusy( TRUE );
}

/*****************************************************************************
 * Handler for the button "Later Tag".
 *****************************************************************************/
void
HandlerToolbarButtonLaterTag( GtkWidget* pWidgetUnused,
                              gpointer   unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return;

   SetSensitivityWhileBusy( FALSE );
   StepHistoryList( 0 );
   SetSensitivityWhileBusy( TRUE );
}

/*****************************************************************************
 * Handler for the button "Goto Tag".
 *****************************************************************************/
void
HandlerToolbarButtonGotoTag( GtkWidget* pWidgetUnused,
                             gpointer   unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return;

   SetSensitivityWhileBusy( FALSE );
   GotoTagUnderCursor();
   SetSensitivityWhileBusy( TRUE );
}

/*****************************************************************************
 * Handler for the button "Record Position".
 *****************************************************************************/
void
HandlerToolbarButtonRecordCursor( GtkWidget* pWidgetUnused,
                                  gpointer   unused )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return;

  /*
   * Is the text widget empty?
   */
   if ( IsTextWidgetEmpty() )
   {
      PrintStatusbar( "Text widget is empty." );
      return;
   }

  /*
   * Record cursor in tag history.
   */
   RecordCursorInTagHistory( "(user-recorded position)" );

   PrintStatusbar( "Recorded cursor position in history." );

   return;
}

/*****************************************************************************
 * Handler for when another notebook page is switched to.
 *
 * HandlerSwitchPageNotebook() and HandlerDrawClistKludge() are part of a kludge,
 * see comments at HandlerDrawClistKludge().
 *****************************************************************************/
void
HandlerSwitchPageNotebook( GtkNotebook*     pNotebook,
                           GtkNotebookPage* pNotebookPage,
                           gint             page_num,
                           gpointer         unused )
{
  /*
   * KLUDGE: Set a flag for tags/modules clist "draw" handler.
   */
   kludgeNotebookSwitchPage = TRUE;

  /*
   * Return void (yes, not FALSE).
   */
   return;
}

/*****************************************************************************
 * KLUDGE:
 * Handler for "draw" signal of tags/modules clists.
 * A known problem is when the notebook page is switched to a tags/modules clist,
 * then sometimes the clists fails to scroll to the selected row, causing
 * the selected row to be outside of the visible area.
 * Kludge is to call gtk_clist_moveto() when the clist receives a "draw" signal
 * just after its notebook page is switched to.
 *****************************************************************************/
void
HandlerDrawClistKludge( GtkWidget*    pWidget,
                        GdkRectangle* pRect,
                        gpointer      unused )
{
   gint row;

  /*
   * (Yes, void will be returned.)
   *
   * If a notebook page is not being switched then don't bother,
   * else clear notebook-page-switch flag and continue.
   */
   if ( kludgeNotebookSwitchPage )
      kludgeNotebookSwitchPage = FALSE;
   else
      return;

  /*
   * Determine which selected-row var to use.
   */
   if ( pWidget == pWidgetClistTags )
      row = widgetTagsClistSelectedRow;
   else if ( pWidget == pWidgetClistModules )
      row = widgetModulesClistSelectedRow;
   else
      return; /* shouldn't happen */

  /*
   * Skip if no row is selected.
   */
   if ( row == ROW_NONE )
      return;

  /*
   * This trick should cause the selected row to appear in the clist scrolled widget.
   */
   gtk_clist_moveto( GTK_CLIST(pWidget),
                     row,
                     0,
                     0.5,
                     0 );

  /*
   * Return void (yes, not FALSE).
   */
   return;
}

/*****************************************************************************
 * Install handlers for when a widget gains/lose focus.
 *****************************************************************************/
void
InstallFocusHandlers( GtkWidget* pWidget )
{
   gtk_signal_connect( GTK_OBJECT(pWidget),
                       "focus-in-event",
                       (GtkSignalFunc)HandlerFocusIn,
                       NULL );
   gtk_signal_connect( GTK_OBJECT(pWidget),
                       "focus-out-event",
                       (GtkSignalFunc)HandlerFocusOut,
                       NULL );
}

/*****************************************************************************
 * Handler for when a widget gains focus.
 *****************************************************************************/
gboolean
HandlerFocusIn( GtkWidget*     pWidget,
                GdkEventFocus* pEvent,
                gpointer       unused )
{
   if ( pWidget == pWidgetText )
   {
      FocusbarFocusGain( pFocusbarText );
   }

  /*
   * Continue to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for when a widget loses focus.
 *****************************************************************************/
gboolean
HandlerFocusOut( GtkWidget*     pWidget,
                 GdkEventFocus* pEvent,
                 gpointer       unused )
{
   if ( pWidget == pWidgetText )
   {
      FocusbarFocusLose( pFocusbarText );
   }

  /*
   * Continue to next handler.
   */
   return FALSE;
}

/*****************************************************************************
 * Handler for an expose event in a focusbar.
 *****************************************************************************/
gboolean
HandlerFocusbarExpose( GtkWidget*      pWidget,
                       GdkEventExpose* pEvent,
                       gpointer        pFocusbar )
{
   FocusbarPaint( (focusbar_t*)pFocusbar );

   return TRUE;
}

/*****************************************************************************
 * User clicked the "Word At Cursor" button in a Find/Text dialog.
 *****************************************************************************/
void
WordAtCursorClicked( void )
{
   GString*		pGstr		= NULL;
   gchar*		pText		= NULL;
   guint		textLen;
   guint		cursorIdx;

  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return;

  /*
   * Freeze text widget, extract text.
   */
   gtk_text_freeze( GTK_TEXT(pWidgetText) );
   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   -1 );
   textLen = gtk_text_get_length( GTK_TEXT(pWidgetText) );
   if ( pText == NULL || textLen > MAX_TEXT_BUFFER_LEN ) goto thaw_abort;

  /*
   * Get index of cursor.
   */
   cursorIdx = TextCursorIdx();
   if ( cursorIdx < 0  ||  cursorIdx > textLen ) goto thaw_abort;

  /*
   * Scoop word at cursor.
   * Return if cursor is not over a word.
   */
   pGstr = ScoopWordAtCursor( MAX_WORD_LEN,
                              pText,
                              textLen,
                              cursorIdx );
   if ( pGstr == NULL )
   {
      goto thaw_quit;
   }

  /*
   * Thaw text widget.
   */
   gtk_text_thaw( GTK_TEXT(pWidgetText) );

  /*
   * Insert scooped word into text entry.
   */
   gtk_entry_set_text( GTK_ENTRY( GTK_COMBO(pWidgetComboFind)->entry ),
                       pGstr->str );

  /*
   * Free mem, return.
   */
quit:
   if ( pText ) g_free( pText );
   if ( pGstr ) g_string_free( pGstr, TRUE );
   return;

  /*
   * Oops.
   */
thaw_abort:
   g_assert( 0 );
thaw_quit:
   gtk_text_thaw( GTK_TEXT(pWidgetText) );
   goto quit;
}

/*****************************************************************************
 * User clicked the "Restore Cursor Position" button in a Find dialog.
 *****************************************************************************/
void
RestoreCursorClicked( void )
{
  /*
   * Prevent nesting major functions.
   */
   if ( IsAnyMajorFunctionBusy() ) return;

   g_assert( buttonRestoreCursorFlag );

  /*
   * Restore state of cursor and text widget before user did a search.
   */
   RestoreCursor( &cursorText );

  /*
   * Close find dialog.
   */
#if 0
      gtk_widget_destroy( pWidgetDialogFind );
#else
      FindDialogClose();
#endif

   return;
}
