/******************************************************************************
 * $Id: sys_linux.c,v 1.3 2000/08/31 03:17:49 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions specific to Linux.
 ******************************************************************************/

#include "header.h"

/*****************************************************************************
 * Returns the amount of CPUs in the system.
 * If an error occurs, silently returns 1 (not 0) ;-)
 *
 * Notes:	This is specific to Linux/GLIB.
 *****************************************************************************/
int
SysCpus( void )
{
   GString*	pGstr;
   gchar*	pTmp;
   static int cpus = 0;

  /*
   * Has this been called before?  If so, amount of CPUs is known.
   */
   if ( cpus ) return cpus;

  /*
   * The only known way to determine how many CPUs are in a Linux box
   * is to parse /proc/cpuinfo.
   */
   pGstr = LoadFileSpecial( "/proc/cpuinfo" );
   if ( pGstr == NULL  ||  pGstr->len <= 0 ) return 1;

  /*
   * Count each occurrence of the str "processor" in /proc/cpuinfo.
   */
   pTmp    = pGstr->str;
   for (;;)
   {
      pTmp = FindStringInSegment( "processor",
                                   pTmp,
                                   pGstr->len - (pTmp - pGstr->str) );
      if ( pTmp )
      {
         ++cpus;
         ++pTmp;
      }
      else
      {
         g_string_free( pGstr, TRUE );
         if ( cpus == 0 ) return 1;		/* don't ever return 0 ;-) */
         return cpus;
      }
   }
}
