/******************************************************************************
 * $Id: text.c,v 1.112 2000/12/21 04:12:20 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Functions for the text widget.
 ******************************************************************************/

#include "header.h"

/******************************************************************************
 * Create the text widget (and creates ancillary widgets such as focusbar, table).
 *
 * Parms   : pWidgetText (OUT/GLOBAL)
 ******************************************************************************/
void
CreateTextWidgetEtAl( void )
{
  /*
   * Create the text widget.
   */
   pWidgetText = gtk_text_new( NULL, NULL );
   gtk_text_set_editable( GTK_TEXT(pWidgetText),
#if 1
/* ### HACK TODO: To show the cursor, text widget must be set editable. ### */
                          TRUE );
#else
                          FALSE );
#endif

  /*
   * Install a keyboard event handler.
   */
   gtk_signal_connect( GTK_OBJECT(pWidgetText),
                       "key_press_event",
                       (GtkSignalFunc)HandlerKeyPressReleaseText,
                       NULL );
   gtk_signal_connect( GTK_OBJECT(pWidgetText),
                       "key_release_event",
                       (GtkSignalFunc)HandlerKeyPressReleaseText,
                       NULL );

  /*
   * Install handlers for when text widget gains/loses focus.
   */
   InstallFocusHandlers( pWidgetText );

  /*
   * Change the tab width only if specified by the command line (or GNOME config).
   * The default tab width is usually best.
   */
   ApplyTabWidth( pWidgetText, tabWidth );

#if 0
/* ## Scrollbars do not work using scrolled windows. ## */
   textScrolled = gtk_scrolled_window_new( NULL, NULL );
   gtk_scrolled_window_set_policy( GTK_SCROLLED_WINDOW(pWidgetTextScrolled),
                                   GTK_POLICY_ALWAYS,
                                   GTK_POLICY_ALWAYS);
   gtk_scrolled_window_add_with_viewport( GTK_SCROLLED_WINDOW(pWidgetTextScrolled),
                                          pWidgetText );
#endif

  /*
   * Change its color.
   */
   ChangeStyleColorText( pWidgetText,
                         &colorTextWidgetTextFg,
                         &colorTextWidgetTextBg );
   ChangeStyleColorSel( pWidgetText,
                        &colorTextWidgetSelFg,
                        &colorTextWidgetSelBg );

  /*
   * Change its font.
   */
   ChangeStyleFont( pWidgetText,
                    fontNameText );

  /*
   * Create a 2x2 table to hold the text widget and its scrollbars.
   */
   pWidgetTextTable = gtk_table_new( 2, 2, FALSE );

  /*
   * Attach the text window to the 2x2 table.
   */
   gtk_table_attach( GTK_TABLE(pWidgetTextTable),
                               pWidgetText,
                               0, 1, 0, 1,
                               GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                               GTK_EXPAND | GTK_SHRINK | GTK_FILL, 0, 0 );

  /*
   * Create the scrollbars for the text widget.
   * Attach the scrollbars to the 2x2 table.
   */
   pWidgetTextHscroll = gtk_hscrollbar_new( GTK_TEXT(pWidgetText)->hadj );
   pWidgetTextVscroll = gtk_vscrollbar_new( GTK_TEXT(pWidgetText)->vadj );
   gtk_table_attach( GTK_TABLE(pWidgetTextTable),
                               pWidgetTextHscroll,
                               0, 1, 1, 2,
                               GTK_EXPAND | GTK_FILL | GTK_SHRINK,
                               GTK_FILL,
                               0, 0);
   gtk_table_attach( GTK_TABLE(pWidgetTextTable),
                               pWidgetTextVscroll,
                               1, 2, 0, 1,
                               GTK_FILL,
                               GTK_EXPAND | GTK_SHRINK | GTK_FILL,
                               0, 0);

  /*
   * Create a "focus bar" to be placed above the text widget.
   */
   pFocusbarText = FocusbarCreate( "8x13" );

  /*
   * Create a 2-slot vbox to hold text widget and its "focus bar".
   */
   pWidgetTextVbox = gtk_vbox_new( FALSE, 0 );
   gtk_box_pack_start( GTK_BOX(pWidgetTextVbox),	/* pack focusbar */
                       pFocusbarText->pWidget,
                       FALSE, FALSE, 0 );
   gtk_box_pack_end( GTK_BOX(pWidgetTextVbox),		/* pack table containing text widget/scrollbars */
                     pWidgetTextTable,
                     TRUE, TRUE, TRUE );

   return;
}

/*****************************************************************************
 * Apply a tab width to the text widget.
 * Does NOT redraw the text widget.
 * Does NOT affect global vars.
 *
 * Parms   : pWidget
 *           Must be a text widget.
 *
 *           width
 *****************************************************************************/
void
ApplyTabWidth( GtkWidget* pWidget,
               int        width )
{
  /*
   * Has text widget been created?
   */
   if ( pWidget == NULL ) return;

  /*
   * width can be 0 (which selects GTK+ default).
   */
   g_assert( width >= 0  &&  width < 30 );

   if ( width > 0 )
   {
     /*
      * Stolen from GXedit ;-)
      */
      GTK_TEXT(pWidget)->default_tab_width = width;
      GTK_TEXT(pWidget)->tab_stops = g_list_remove( GTK_TEXT(pWidget)->tab_stops,
                                                    GTK_TEXT(pWidget)->tab_stops->data );
      GTK_TEXT(pWidget)->tab_stops = NULL;
      GTK_TEXT(pWidget)->tab_stops = g_list_prepend( GTK_TEXT(pWidget)->tab_stops,
                                                     (gpointer)tabWidth );
   }
}

/******************************************************************************
 * Load the text widget with a module (ie source file).
 * Also moves the cursor to the specified line.
 * Frees buffer holding previous module depending on discardTextBuffers.
 *
 * Returns : FUNC_OK or FUNC_FAIL
 *
 * Parms   : pModule
 *           Name of module (typically a file name).
 *
 *         : cursor
 *           Which line to move the cursor to.
 *
 *         : row
 *           The associated list widget row (use ROW_NONE if N/A).
 *
 *         : moduleLoadedIntoTextWidget (GLOBAL)
 *           Assigned with the name of the module that is loaded
 *           into the text widget.
 *
 *         : pHashModules (GLOBAL)
 *           The module hash will be updated if the text widget is copied out
 *           (into a new buffer) before being loaded with another module.
 *
 *         : reload
 *           If TRUE then the text widget will be loaded, even if it already
 *           contains the requested module.
 ******************************************************************************/
FUNC_X
LoadTextWidgetWithModule( gchar*   pModule,
                          gint	   cursor,
                          gint     row,
                          gboolean reload )
{
   gpointer		pBuffer = NULL;
   long			size;
   const char*	pLine = NULL;

  /*
   * Load the text widget only if it does not already contain the specified module
   * (or a reload is requested).
   * strcmp() returns !0 if strings differ.
   */
   if ( reload  ||  strcmp( pModule, moduleLoadedIntoTextWidget ) )
   {
     /*
      * Without a known way to determine that the text widget has been modified,
      * always copy out the text widget.
      * (Does gtk_editable_changed() indicate a modified text widget?)
      */
      CopyOutTextWidget();

     /*
      * Discard previous text buffer (if configured so).
      */
      if ( discardTextBuffers )
         FreeModuleBuffer( moduleLoadedIntoTextWidget );

     /*
      * Obtain a pointer to the buffer that holds the specified module
      * (loading it from a file if necessary).
      */
      pBuffer = ModuleBuffer( pModule,
                              &size /* out */ );
      if ( ! pBuffer ) return FUNC_FAIL;

     /*
      * Actually load the text widget with the module buffer.
      */
      LoadTextWidgetWithBuffer( pBuffer, size );

     /*
      * Record the name of the module loaded into the text widget.
      */
      strcpy( moduleLoadedIntoTextWidget, pModule );
   }
   else
   {
     /*
      * Text widget already contains the requested module.
      * Just get a pointer to its buffer (for the following).
      */
      pBuffer = ModuleBuffer( pModule,
                              &size /* out */ );
      if ( ! pBuffer ) return FUNC_FAIL;
   }

  /*
   * Move the cursor to the specified line.
   */
   pLine = ToLine( pBuffer, cursor, MAX_LINE_LEN );
   if ( ! pLine )
   {
     /*
      * Unable to find this line.
      */
      PrintStatusbarForRow( "This line is beyond the end.  Tag information probably is stale.",
                            row );
      return FUNC_FAIL;
   }
   ScrollTextWidgetToLine( cursor,
                           SCROLL_TEXT_VERT_FRAC );

  /*
   * Update the text widget's focusbar label, to show which module is being shown.
   */
   FocusbarLabelSet( pFocusbarText,
                     pModule );

   return FUNC_OK;
}

/******************************************************************************
 * Load the text widget with a memory buffer (overwriting current contents).
 *
 * (This function is generic, therefore, it is ignorant of what the buffer is.)
 *
 * Parms   : pBuffer
 *
 *         : bufferSize
 ******************************************************************************/
void
LoadTextWidgetWithBuffer( gpointer pBuffer,
                          gint     bufferSize )
{
   gtk_text_freeze( GTK_TEXT(pWidgetText) );

  /*
   * Clear the text widget.
   */
   ClearTextWidget();

  /*
   * Insert the buffer into the text widget.
   */
   if ( hilite )
     /*
      * TODO call should be based on file ext (eg .c,.h,etc)
      */
      HiliteLoadCCode( pWidgetText,
                       pBuffer,
                       bufferSize );
   else
      gtk_text_insert( GTK_TEXT(pWidgetText),
                       NULL,
                       NULL,
                       NULL,
                       pBuffer,
                       bufferSize );

   gtk_text_thaw( GTK_TEXT(pWidgetText) );

   return;
}

/******************************************************************************
 * Copy out the contents of text widget (which holds a module) into a new buffer.
 *
 * The buffer will be appended with a trailing null char (unaccounted by size)
 * to facilitate processing by string funcs.
 *
 * Doing so prepares the text widget for being reloaded.
 * Does nothing if the text widget has no module (ie is empty).
 * The hash item (corresponding to the outgoing module) will be updated.
 *
 * Parms   : moduleLoadedIntoTextWidget (GLOBAL)
 *           The name of the module currently within the text widget.
 *           Not changed here.
 *
 *         : pHashModules (OUT/GLOBAL)
 *           The module hash will be updated.
 ******************************************************************************/
void
CopyOutTextWidget( void )
{
   P_HASH_ITEM_MODULE	pHashItem;
   char*				pText		= NULL;

  /*
   * Do nothing if no module has been loaded into the text widget.
   */
  if ( moduleLoadedIntoTextWidget[0] == '\0' ) return;

  /*
   * Find the hash item that corresponds to the module currently in the text widget.
   */
   pHashItem = \
   (P_HASH_ITEM_MODULE)g_hash_table_lookup( pHashModules,
                                            (gconstpointer)moduleLoadedIntoTextWidget );
g_assert( pHashItem );

  /*
   * Free the old buffer.
   * If discardTextBuffers is true then it might already be freed,
   * but otherwise it should still be allocated.
   */
if ( ! discardTextBuffers ) g_assert( pHashItem->pModuleBuffer );
   FreeModuleBuffer( moduleLoadedIntoTextWidget );

  /*
   * To copy out the contents of the text widget, call gtk_editable_get_chars(),
   * which returns a newly allocated buffer.
   */
   pHashItem->moduleSize = gtk_text_get_length( GTK_TEXT(pWidgetText) );
   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   pHashItem->moduleSize );
g_assert( pText  &&  pHashItem->moduleSize );

  /*
   * Create a new buffer with the contents of the text widget,
   * plus a trailing null char (not accounted by size).
   */
   pHashItem->pModuleBuffer = g_malloc( pHashItem->moduleSize + 1 );	/* alloc buffer */
g_assert( pHashItem->pModuleBuffer );
   memcpy( pHashItem->pModuleBuffer, pText, pHashItem->moduleSize );	/* copy buffer */
   ((char*)pHashItem->pModuleBuffer)[pHashItem->moduleSize] = '\0';		/* append null char */

  /*
   * Free memory from gtk_editable_get_chars().
   */
   g_free( pText );

   return;
}

/******************************************************************************
 * Clear the text widget.
 *
 * Precond : The caller should first freeze the text widget.
 ******************************************************************************/
void
ClearTextWidget( void )
{
  /*
   * First, move the cursor to the first char.
   */
   gtk_text_set_point( GTK_TEXT(pWidgetText), 0 );

  /*
   * Now, delete all following chars.
   */
   gtk_text_forward_delete( GTK_TEXT(pWidgetText),
                            gtk_text_get_length( GTK_TEXT(pWidgetText) ) );

   return;
}

/*****************************************************************************
 * Redraw the text widget (unless empty).
 *****************************************************************************/
void
RedrawTextWidget( void )
{
   int		line;
   gchar	moduleThatWasLoaded[MAX_FILENAME_LEN];

   if ( ! IsTextWidgetEmpty() )
   {
      line = LineAtCursor();
      strcpy( moduleThatWasLoaded, moduleLoadedIntoTextWidget );
      LoadTextWidgetWithModule( moduleThatWasLoaded,
                                line,
                                ROW_NONE,
                                DO_RELOAD );
   }
}

/******************************************************************************
 * Return true if the text widget is empty.
 ******************************************************************************/
int
IsTextWidgetEmpty( void )
{
   return gtk_text_get_length( GTK_TEXT(pWidgetText) ) == 0;
}

/******************************************************************************
 * Return the line where the cursor is at in the text widget.
 * Returns -1 upon failure.
 ******************************************************************************/
int
LineAtCursor( void )
{
   gchar*	pText = NULL;
   int		line;

   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   -1 );
   if ( ! pText ) return -1;

#if 0
   line = LineIdx2LineNum( pText,
                           gtk_text_get_point( GTK_TEXT(pWidgetText) ) );
#else
   line = LineIdx2LineNum( pText,
                           gtk_editable_get_position( GTK_EDITABLE(pWidgetText) ) );
#endif

   g_free( pText );

   return line;
}

/******************************************************************************
 * Scroll the text widget so that a line is visible.
 *
 * This code was stolen from GXedit ;-)
 *
 * BUG TODO: If the text has few lines and "line" is passed as zero,
 *           this function will incorrectly scroll downward.
 *
 * Parms   : line
 *           Which line # to scroll to.
 *
 *           vertFrac
 *           {0.0 ... 1.0} in which 0.0 scrolls line to top, 1.0 to bottom.
 ******************************************************************************/
void
ScrollTextWidgetToLine( int   line,
                        float vertFrac )
{
   int		i;
   int		cl = 0;
   int		tl = 0;
   int		text_len;
   char*	alltext = NULL;

  /*
   * line == 0 causes gtk_text_insert() to erratically assert.
   */
   if ( line == 0 ) line = 1;

  /*
   * Freeze text widget, extract text.
   */
   gtk_text_freeze( GTK_TEXT(pWidgetText) );
   text_len = gtk_text_get_length( GTK_TEXT(pWidgetText) );
   if ( text_len == 0 ) goto thaw_abort;
   alltext = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText), 0, text_len );
   if ( alltext == NULL ) goto thaw_abort;

   for ( i = 0; i < text_len; i++ )
   {
      if ( IsEol( &alltext[i] ) ) tl++;
   }

   for ( i = 0; i < text_len; i++ )
   {
      if ( IsEol( &alltext[i] ) ) cl++;
      if ( cl == line ) break;
   }
   ++i;

  /*
   * Overshot?
   * If this is not checked then GTK will abort program!!!
   */
   if ( i > text_len ) goto thaw_quit;

  /*
   * Now, thaw text widget, to let it scroll.
   */
   gtk_text_thaw( GTK_TEXT(pWidgetText) );

  /*
   * Skip setting an scroll increment (ie adjustment) value if the text widget
   * has only one page, otherwise, the unnecessary scroll is annoying to user.
   * One page is indicated if upper == pageinc (multiple pages if upper > paginc).
   */
   if ( GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj)->upper > GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj)->page_increment )
   {
     /*
      * A 0.5 * page_increment is what causes the scrolled-to-line
      * to be positioned at the middle of the text widget.
      */
      gtk_adjustment_set_value( GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj),
                                ((cl * GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj)->upper) / tl - vertFrac * GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj)->page_increment) );
   }
   gtk_editable_insert_text( GTK_EDITABLE(pWidgetText),
                             " ",
                             1,
                             &i );
   gtk_editable_delete_text( GTK_EDITABLE(pWidgetText),
                             i-1,
                             i );

  /*
   * Without this, when trying to highlight the previous line,
   * the preceding line can be highlighted instead.
   */
#if 0
   gtk_text_set_point( GTK_TEXT(pWidgetText),
                       1 + gtk_text_get_point( GTK_TEXT(pWidgetText) ) );
#endif
#if 0
   gtk_editable_set_position( GTK_EDITABLE(pWidgetText),
                              1 + gtk_editable_get_position( GTK_EDITABLE(pWidgetText) ) );
#endif

  /*
   * Text widget has been thawed.
   * Note that re-thawing causes problems (disappearing text cursor).
   */
quit:
   if ( alltext ) g_free( alltext );
   return;

thaw_quit:
   gtk_text_thaw( GTK_TEXT(pWidgetText) );
   goto quit;

thaw_abort:
#if 0
   goto thaw_quit;
#else
   g_assert( 0 );
#endif
}

/******************************************************************************
 * Ie emulate emacs's Ctrl+L.
 ******************************************************************************/
void
RecenterTextWidget( void )
{
   guint cursor;

  /*
   * ScrollTextWidgetToLine() will do so by passing the current line #.
   */
   if ( ! IsTextWidgetEmpty()  &&  LineAtCursor() > -1 )
   {
      cursor = gtk_editable_get_position( GTK_EDITABLE(pWidgetText) );
      ScrollTextWidgetToLine( LineAtCursor(),
                              0.5 );
      gtk_editable_set_position( GTK_EDITABLE(pWidgetText), cursor );
   }
   return;
}

/*****************************************************************************
 * Create/show a dialog widget for jumping to a line # entered by user.
 * This is called when the user selects the menu item /Goto/Line.
 *
 * Parms   : pWidgetDialogLine (OUT/GLOBAL)
 *           Pointer to the created dialog widget for /Goto/Line.
 *****************************************************************************/
void
GotoLine( void )
{
   GtkWidget*	pWidgetHbox;
   GtkWidget*	pWidgetLabel;

  /*
   * Is the text widget empty?
   */
   if ( IsTextWidgetEmpty() )
   {
      PrintStatusbar( "Text widget is empty." );
      return;
   }

  /*
   * Create the dialog widget.
   */
   pWidgetDialogLine = gtk_dialog_new();

  /*
   * Install a key press event handler (to close dialog if ESC is pressed).
   */
   gtk_signal_connect( GTK_OBJECT(pWidgetDialogLine),
                       "key_press_event",
                       (GtkSignalFunc)HandlerKeyPressDialogLine,
                       NULL );

  /*
   * Title the dialog.
   */
   gtk_window_set_title( GTK_WINDOW(pWidgetDialogLine),
                         "Goto Line" );

  /*
   * Widen the border of the dialog.
   */
   gtk_container_border_width( GTK_CONTAINER(pWidgetDialogLine),
                               16 );

  /*
   * Create a hbox to hold the widgets of the dialog.
   */
   pWidgetHbox = gtk_hbox_new( FALSE, 8 );

  /*
   * Create/pack a label for the dialog.
   */
   pWidgetLabel = gtk_label_new( "Line: " );
   gtk_box_pack_start( GTK_BOX(pWidgetHbox),
                       pWidgetLabel,
                       TRUE, TRUE, 0 );

  /*
   * Create/pack the combo (text entry field) of the dialog.
   * Do not add pop-down strings if this is the first instance.
   * gtk_combo_disable_active() makes Enter key work as expected.
   */
   pWidgetComboLine = gtk_combo_new();
   gtk_combo_disable_activate( GTK_COMBO(pWidgetComboLine) );
   if ( pGlistLineStrings != NULL )
      gtk_combo_set_popdown_strings( GTK_COMBO(pWidgetComboLine),
                                     pGlistLineStrings );
   gtk_box_pack_start( GTK_BOX(pWidgetHbox),
                       pWidgetComboLine,
                       TRUE, TRUE, 0 );

  /*
   * Clear the text entry widget (it repeats the last item typed).
   */
   gtk_entry_set_text( GTK_ENTRY(GTK_COMBO(pWidgetComboLine)->entry),
                       "" );

  /*
   * The final pack is to pack the hbox into the dialog's vbox.
   */
   gtk_box_pack_start( GTK_BOX(GTK_DIALOG(pWidgetDialogLine)->vbox),
                       pWidgetHbox,
                       TRUE, TRUE, 0 );

  /*
   * Connect to signals.
   */
   gtk_signal_connect( GTK_OBJECT(GTK_COMBO(pWidgetComboLine)->entry),
                       "activate",
                       GTK_SIGNAL_FUNC(HandlerActivateComboLine),
                       pWidgetDialogLine );

  /*
   * Position/show the dialog (and its sub-widgets).
   */
   PositionShowDialogAtMiddle( pWidgetDialogLine );

  /*
   * Grab the focus for the entry widget.
   */
   gtk_widget_grab_focus( GTK_WIDGET(GTK_COMBO(pWidgetComboLine)->entry) );

   return;
}

/******************************************************************************
 * Highlight a line in the text widget.
 *
 * Returns false if line is invalid.
 ******************************************************************************/
gboolean
HighlightLine( int line )
{
   char*	pText		= NULL;
   char*	pLine;
   char*	pLineEnd;

  /*
   * Extract contents of text widget.
   */
   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   -1 );
   if ( pText == NULL ) goto fail;

  /*
   * Compute pointer to specified line.
   */
   pLine = (char*)ToLine( pText,
                          line,
                          MAX_LINE_LEN );
   if ( pLine == NULL ) goto fail;

  /*
   * Compute pointer to end of specified line.
   */
   pLineEnd = ToEol( pLine, MAX_LINE_LEN ) - 1;

  /*
   * Return if NULLs.
   */
   if ( pLine == NULL  ||  pLineEnd == NULL )
      goto fail;

   gtk_editable_select_region( GTK_EDITABLE(pWidgetText),
                               pLine    - pText,
                               pLineEnd - pText );

   g_free( pText );
   return TRUE;

fail:
   g_free( pText );
   return FALSE;
}

/******************************************************************************
 * Highlight the first matching word in the current line of the text widget.
 * If unable to find a match, as a contigency, highlight the entire line.
 *
 * GTK's point must point to the first char of the current line,
 * as set by gtk_text_set_point().
 ******************************************************************************/
void
HighlightWordInCurrentLine( const char* pWord )
{
   int		i;
   int		end;
   gchar*	pLine = NULL;			/* holds extracted line */
   int		idxLineStart;			/* index to start of current line in text widget */
   int		idxLineEndTentative;	/* likely will index past the end, TBD */
   int		found = FALSE;

  /*
   * Extract enough from the text widget to include the current line.
   * The end index (which is guessed) must be chosen so that it is not beyond
   * the end of the text widget, otherwise, gtk_editable_get_chars() will fail.
   */
   idxLineStart        = gtk_text_get_point( GTK_TEXT(pWidgetText) );
   idxLineEndTentative = idxLineStart + MAX_LINE_LEN;
   if ( idxLineEndTentative >= gtk_text_get_length( GTK_TEXT(pWidgetText) ) )
      idxLineEndTentative = gtk_text_get_length( GTK_TEXT(pWidgetText) ) - 1;
   pLine = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   idxLineStart,
                                   idxLineEndTentative );
   if ( pLine == NULL ) goto quit;

  /*
   * Find the index of the first matching word in the line.
   * The loop ends when the last char of the sought word
   * would hit the tentative end of the line.
   */
   end = idxLineEndTentative - idxLineStart - strlen( pWord );
   if ( end < 0 || end > MAX_LINE_LEN ) { /* Oops, bail out. */ goto quit; }
   for ( i = 0; i <= end; ++i )
   {
      if ( strncmp( &pLine[i], pWord, strlen( pWord ) ) == 0 )
      {
         found = TRUE;
         break;
      }
   }

  /*
   * Highlight the first match (if found).
   */
   if ( found )
   {
     /*
      * Move cursor to start of highlight before highlighting.
      */
      gtk_editable_set_position( GTK_EDITABLE(pWidgetText),
                                 idxLineStart + i );
      ProcessPendingEvents(); /* otherwise selection will sometimes disappear */
      gtk_editable_select_region( GTK_EDITABLE(pWidgetText),
                                  idxLineStart + i,
                                  idxLineStart + i + strlen( pWord ) );
      goto quit;
   }

  /*
   * No match.  As a contigency plan, highlight the entire line.
   * Now find the exact end of the current line.
   */
   for ( i = 0; i <= (idxLineEndTentative - idxLineStart); ++i )
   {
      if ( pLine[i] == '\0'  ||  IsEol( &pLine[i] ) )
      {
         gtk_editable_select_region( GTK_EDITABLE(pWidgetText),
                                     idxLineStart,
                                     idxLineStart + i );
         break;
      }
   }

quit:
   g_free( pLine );
   return;
}

/*****************************************************************************
 * Highlight a string in the text widget.
 *
 * The text widget will be scrolled to show the string.
 *
 * Parms   : pText
 *           Pointer to the first char of the text of the text widget.
 *
 *           idx
 *           Index of the string to highlight.
 *
 *           len
 *           Length of string to highlight.
 *****************************************************************************/
void
HighlightTextString( gchar* pText,
                     gint   idx,
                     gint   len )
{
   g_return_if_fail( pText );
   g_return_if_fail( idx >= 0  &&  idx <= MAX_LINE_LEN * MAX_LINES );
   g_return_if_fail( len >  0  &&  len <= MAX_LINE_LEN );

  /*
   * Scroll the text widget to the string.
   */
   ScrollTextWidgetToLine( LineIdx2LineNum( pText, idx ),
                           SCROLL_TEXT_VERT_FRAC );

  /*
   * Move cursor to start of highlight.
   */
   gtk_editable_set_position( GTK_EDITABLE(pWidgetText),
                              idx );
   ProcessPendingEvents(); /* otherwise selection will sometimes disappear */
  /*
   * Highlight the string.
   */
   gtk_editable_select_region( GTK_EDITABLE(pWidgetText),
                               idx,
                               idx + len );
   return;
}

/*****************************************************************************
 * Highlight the tag in the text widget.
 *****************************************************************************/
void
HighlightTag( struct TAG* pTag )
{
  /*
   * Highlight the tag in the text widget.
   * ctags inserts a space between "operatorX",
   * so to ensure the highlight search works, just specify "operator".
   */
   if ( ! strstr( pTag->pName, "operator" ) )
      if ( ! strstr( pTag->pName, "(recorded position)" ) )
         HighlightWordInCurrentLine( pTag->pName );
      else
         HighlightLine( pTag->lineInt );
   else
      HighlightWordInCurrentLine( "operator" );

   return;
}

/*****************************************************************************
 * Print the line # in the statusbar.
 *****************************************************************************/
void
PrintLineNum( void )
{
   char msg[20];

  /*
   * Skip if text widget is empty.
   */
   if ( IsTextWidgetEmpty() ) return;

  /*
   * Print the line # in the statusbar.
   */
/* !!! Change PrintStatusbar() if the below parm #2 is changed !!! */
   sprintf( msg, " line: %d", LineAtCursor()+1 );
   PrintStatusbar( msg );

   return;
}

/*****************************************************************************
 * Prints the line # in the statusbar.
 * Called by GTK+ when GTK+ is idle, then shuts off.
 * Does nothing if the text widget is empty.
 *****************************************************************************/
gint
IdlePrintLineNum( gpointer dontCare )
{
  /*
   * Only print line # if text widget is not empty.
   */
   if ( ! IsTextWidgetEmpty() )
      PrintLineNum();

  /*
   * Returning zero causes GTK+ to stop calling this idle func.
   */
   return 0;
}

/*****************************************************************************
 * Return index of cursor of text widget.
 * Caller should ensure text widget is not empty!
 *****************************************************************************/
guint
TextCursorIdx( void )
{
g_assert( ! IsTextWidgetEmpty() );

#if 1
   return gtk_editable_get_position( GTK_EDITABLE(pWidgetText) );
#else
   return gtk_text_get_point( GTK_TEXT(pWidgetText) );
#endif
}

/******************************************************************************
 * Return true if cursor is over a word in the text widget.
 ******************************************************************************/
gboolean
IsCursorOverWord( void )
{
   gchar*	pText	= NULL;
   gboolean	res;

   if ( IsTextWidgetEmpty() ) return FALSE;

  /*
   * Extract text from the text widget.
   */
   gtk_text_freeze( GTK_TEXT(pWidgetText) );
   pText = gtk_editable_get_chars( GTK_EDITABLE(pWidgetText),
                                   0,
                                   -1 );
   gtk_text_thaw( GTK_TEXT(pWidgetText) );
   g_return_val_if_fail( pText, FALSE );

  /*
   * Return whether cursor is over a char which is/isnt part of a word.
   */
   res = IsTagChar( pText[gtk_editable_get_position( GTK_EDITABLE(pWidgetText))] );
   g_free( pText );
   return res;
}

/*****************************************************************************
 * Scoop the word which the cursor is over (in text widget).
 *
 * Parms   : maxWordLen
 *           Returns failure if word is larger.
 *
 *           pText
 *           Pointer to extracted data from text widget.
 *
 *           textLen
 *           Length of extracted data from text widget.
 *
 *           idxCursor
 *           Index of cursor
 *
 * Returns : Pointer to a GString, or NULL.
 *****************************************************************************/
GString*
ScoopWordAtCursor( int   maxWordLen,
                   char* pText,
                   int   textLen,
                   int   idxCursor )
{
   GString* pGstr	= NULL;
   int		idxWordStart;
   int		idxWordEnd;
   int		i;

   g_assert(    pText
             && textLen > 0
             && textLen < MAX_TEXT_BUFFER_LEN
             && idxCursor >= 0
             && idxCursor <= textLen );		/* GTK lets cursor idx equal text len */

   idxWordStart = idxCursor;

  /*
   * Is the cursor is over a non-word char (ie whitespace)?
   */
   if ( ! IsWordChar( pText[idxWordStart] ) )
      return NULL;

  /*
   * Compute idxWordStart.
   *
   * If the cursor looks like it is over the middle of a word,
   * then scan backwards for the start of the word.
   */
   for ( i = 0; i < maxWordLen; ++i )
   {
     /*
      * To next preceding char.
      */
      --idxWordStart;

     /*
      * Break if encountered the first preceding non-word char,
      * or beginning of text widget.
      */
      if (    (idxWordStart < 0)
           || (! IsWordChar( pText[idxWordStart] )) )
      {
        /*
         * Move index forward to start of word, break.
         */
         ++idxWordStart;
         break;
      }

     /*
      * Index is still at a word char.
      * Break if the index is now 0 (word is at beginning of text widget).
      */
      if ( idxWordStart == 0 ) break;
   }
   if ( i >= maxWordLen ) return NULL;

  /*
   * Compute idxWordEnd.
   */
   idxWordEnd = idxWordStart + 1;
   for ( i = 0; i < maxWordLen; ++i )
   {
     /*
      * Break if end of text reached.
      */
      if ( idxWordEnd >= textLen )
      {
         idxWordEnd = textLen - 1;
         break;
      }

     /*
      * Scan until the first non-word char that trails the word.
      */
      if ( ! IsWordChar( pText[idxWordEnd] ) )
      {
         --idxWordEnd;
         break;
      }

     /*
      * Forward to next char.
      */
      ++idxWordEnd;
   }
   if ( i >= maxWordLen ) return NULL;

  /*
   * Abort if the indexes were somehow miscomputed.
   */
   g_assert(    idxWordEnd >= idxWordStart
             && (idxWordEnd - idxWordStart + 1) > 0
             && (idxWordEnd - idxWordStart + 1) <= maxWordLen );

  /*
   * Copy the word into a GString.
   */
   pGstr = CreateGString( &pText[idxWordStart],
                          &pText[idxWordEnd] - &pText[idxWordStart] + 1 );

  /*
   * Return GString.
   */
   return pGstr;
}

/*****************************************************************************
 * Record the current position in the text widget.
 *
 * Parms   : pCursor (OUT)
 *           Pointer to a struct that defines state of the cursor position.
 *****************************************************************************/
void
SaveCursor( CURSOR* pCursor )
{
  /*
   * If text widget is empty then invalidate struct, silently return.
   */
   if ( IsTextWidgetEmpty() )
   {
      pCursor->valid = FALSE;
      return;
   }

  /*
   * Fill the passed CURSOR struct.
   */
   pCursor->valid           = TRUE;
   pCursor->adjustmentValue = GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj)->value;
   pCursor->idxCursor       = TextCursorIdx();
   strncpy( pCursor->module, moduleLoadedIntoTextWidget, MAX_FILENAME_LEN );

   return;
}

/*****************************************************************************
 * Restore the current position in the text widget.
 *
 * Parms   : pCursor (IN)
 *           Pointer to a struct that defines state of the cursor position.
 *****************************************************************************/
void
RestoreCursor( CURSOR* pCursor )
{
  /*
   * Quit if CURSOR struct is not valid
   * (if so, text widget was empty when SaveCursor() was called)
   */
   if ( ! pCursor->valid ) return;

  /*
   * This func will only load a module if the text widget
   * is not already holding it.
   */
   LoadTextWidgetWithModule( pCursor->module,
                             0,
                             ROW_NONE,
                             DONT_RELOAD );

  /*
   * Restore the cursor position and the state of the text widget.
   */
   gtk_adjustment_set_value( GTK_ADJUSTMENT(GTK_TEXT(pWidgetText)->vadj),
                             pCursor->adjustmentValue );
   gtk_editable_set_position( GTK_EDITABLE(pWidgetText),
                              pCursor->idxCursor );

   return;
}
