/******************************************************************************
 * $Id: widgets.c,v 1.112 2001/01/01 00:10:00 jimb Exp $
 * Author		: Jim Brooks	<jim@jimbrooks.org>
 * Description	: Other widget functions.
 ******************************************************************************/

#include "header.h"

#include <gdk/gdk.h>
#include <gdk/gdkkeysyms.h>

/******************************************************************************
 * Create all of the widgets.
 *
 * Parms   : argc, argv
 *
 *           many pWidgetsX (OUT/GLOBAL)
 ******************************************************************************/
void
CreateWidgets( void )
{
  /*
   * Create the main window.
   */
   CreateMainWindow();
   if ( mainWindowTitle[0] )
      gtk_window_set_title( GTK_WINDOW(pWidgetMain), mainWindowTitle );

  /*
   * Create a vbox and attach it to the main window.
   */
   pWidgetVbox = gtk_vbox_new( FALSE, 0 );
   gtk_container_add( GTK_CONTAINER(pWidgetMain),
                      pWidgetVbox );

  /*
   * Create the menubar/toolbar.
   */
   CreateMenubarToolbar();

   SetSensitivityWhileBusy( FALSE );

  /*
   * Create the tags clist widget.
   */
   CreateTagsClistWidget();

  /*
   * Create the modules clist widget.
   */
   CreateModulesClistWidget();

  /*
   * Create the history clist widget.
   */
   CreateHistoryClistWidget();

  /*
   * Create the notebook widget (to be added to the left half of the paned widget).
   */
   CreateNotebook();

  /*
   * Create the text widget et al (to be added to the right half of the paned widget).
   */
   CreateTextWidgetEtAl();

  /*
   * Create the paned widget (which contains the notebook and text widgets).
   */
   pWidgetPaned = gtk_hpaned_new();
   gtk_box_pack_start( GTK_BOX(pWidgetVbox),
                       pWidgetPaned,
                       TRUE, TRUE, TRUE);
   textLeftApplied = textLeft; /* remember how layout was applied */
   if ( textLeft == FALSE )
   {
      gtk_paned_add1( GTK_PANED(pWidgetPaned), pWidgetNotebook );
      gtk_paned_add2( GTK_PANED(pWidgetPaned), pWidgetTextVbox );
   }
   else
   {
      gtk_paned_add2( GTK_PANED(pWidgetPaned), pWidgetNotebook );
      gtk_paned_add1( GTK_PANED(pWidgetPaned), pWidgetTextVbox );
   }
   gtk_widget_set_usize( GTK_WIDGET(pWidgetPaned),
                         PANED_WIDTH_HORZ,
                         PANED_WIDTH_VERT );

  /*
   * Create the status bar.
   */
   pWidgetStatusbar = gtk_statusbar_new();
   gtk_box_pack_end( GTK_BOX(pWidgetVbox),
                     pWidgetStatusbar,
                     0, 0, 0);
   statusbarContextId = gtk_statusbar_get_context_id( GTK_STATUSBAR( pWidgetStatusbar ),
                                                      "hypersrc" );

  /*
   * Show all windows.
   */
   gtk_widget_show_all( pWidgetMain );

  /*
   * Some widget operations can only be done after a widget is visible.
   */
   FocusbarLabelSet( pFocusbarText, "(empty)" );

  /*
   * Position the paned slider.
   */
   SetPanedSliderInitialPosition();

  /*
   * Populate the tags clist widget.
   */
   PopulateTagsClistWidget( SORT_MODULE );

  /*
   * Populate the modules clist widget.
   */
   PopulateModulesClistWidget();

  /*
   * Print a banner in status bar (unless the status bar has a pending msg).
   */
   SetStatusbarInitialMsg( HYPERSRC_TITLE_STRING );
   PrintStatusbar( pStatusbarInitMsg );

  /*
   * Install a key press event handler.
   */
   gtk_signal_connect( GTK_OBJECT (pWidgetMain),
                       "key-press-event",
                       (GtkSignalFunc)HandlerKeyPressMain,
                       NULL );

  /*
   * Enable a handler for updating the line #.
   */
#if 0
/* HANDLER WILL NEVER BE CALLED */
/* LEFT HERE AS INFO */
   gtk_signal_connect( GTK_OBJECT(pWidgetText),
                       "move_cursor",
                       (GtkSignalFunc)HandlerLineNum,
                       NULL );
#endif
   gtk_signal_connect( GTK_OBJECT(pWidgetText),
                       "button-press-event",
                       (GtkSignalFunc)HandlerLineNum,
                       NULL );
   gtk_signal_connect( GTK_OBJECT(pWidgetText),
                       "key-press-event",
                       (GtkSignalFunc)HandlerLineNum,
                       NULL );
   gtk_signal_connect( GTK_OBJECT(pWidgetText),
                       "changed",
                       (GtkSignalFunc)HandlerLineNum,
                       NULL );

   SetSensitivityWhileBusy( TRUE );

  /*
   * Enter message loop.
   */
   gtk_main();
}

/******************************************************************************
 * Create the main window.
 *
 * The main window can be created using GTK+, rather than gnome_app_new().
 *
 * Parms   : pWidgetMain (OUT/GLOBAL)
 ******************************************************************************/
void
CreateMainWindow( void )
{
  /*
   * Create the main window.
   */
   pWidgetMain = gtk_window_new( GTK_WINDOW_TOPLEVEL );
   gtk_signal_connect( GTK_OBJECT(pWidgetMain),
                       "destroy",
                       GTK_SIGNAL_FUNC(gtk_main_quit),
                       NULL );
   gtk_window_set_title( GTK_WINDOW (pWidgetMain),
                         "hypersrc" );

  /*
   * If "-geom" was specified (indicated by widgetMainGeom[0] != -1),
   * then change the geometry of the main widget.
   */
   if ( widgetMainGeom[0] != -1 )
   {
      gtk_widget_set_uposition( GTK_WIDGET(pWidgetMain),
                                widgetMainGeom[0],
                                widgetMainGeom[1] );
#if 0
    /*
     * This forces minimum size of widget.
     */
      gtk_widget_set_usize( GTK_WIDGET(pWidgetMain),
                            widgetMainGeom[2],
                            widgetMainGeom[3] );
#else
      gtk_window_set_default_size( GTK_WINDOW(pWidgetMain),
                                   widgetMainGeom[2],
                                   widgetMainGeom[3] );
#endif
   }

   return;
}

/******************************************************************************
 * Create the menubar/toolbar within a framed hbox, then attach to main vbox.
 ******************************************************************************/
void
CreateMenubarToolbar( void )
{
   GtkWidget*	pWidgetHboxMenubarToolbar;
   GtkWidget*	pWidgetFrameMenubarToolbar;

  /*
   * Create the menubar and toolbar within their own hbox.
   */
   pWidgetHboxMenubarToolbar = gtk_hbox_new( FALSE, 0 );
   CreateMenubar( pWidgetHboxMenubarToolbar );
   CreateToolbar( pWidgetHboxMenubarToolbar );

  /*
   * Frame the menu/toolbar hbox.
   */
   pWidgetFrameMenubarToolbar = gtk_frame_new( NULL );
   gtk_frame_set_label_align( GTK_FRAME(pWidgetFrameMenubarToolbar),
                              0.10, 0 );
   gtk_frame_set_shadow_type( GTK_FRAME(pWidgetFrameMenubarToolbar),
                              GTK_SHADOW_OUT );
   gtk_container_add( GTK_CONTAINER(pWidgetFrameMenubarToolbar),
                      pWidgetHboxMenubarToolbar );

  /*
   * Pack the menu/toolbar hbox into the main vbox.
   */
   gtk_box_pack_start( GTK_BOX(pWidgetVbox),
                       pWidgetFrameMenubarToolbar,
                       FALSE, FALSE, 0 );

   return;
}

/******************************************************************************
 * Change the color of text of a widget.
 ******************************************************************************/
void
ChangeStyleColorText( GtkWidget* pWidget,
                      GdkColor*  pColorFg,
                      GdkColor*  pColorBg )
{
   GtkStyle*	pStyle;

  /*
   * Change the color of the widget indirectly
   * by changing the color in the style struct.
   */
   pStyle = gtk_widget_get_style( pWidget );
   pStyle = gtk_style_copy( pStyle );

   pStyle->fg[GTK_STATE_NORMAL]   = *pColorFg;	/* applies to clist fg */
   pStyle->text[GTK_STATE_NORMAL] = *pColorFg;	/* applies to text  fg */
   pStyle->base[GTK_STATE_NORMAL] = *pColorBg;

   gtk_widget_set_style( pWidget, pStyle );

   return;
}

/******************************************************************************
 * Change the color of a selection of a widget.
 ******************************************************************************/
void
ChangeStyleColorSel( GtkWidget* pWidget,
                     GdkColor*  pColorFg,
                     GdkColor*  pColorBg )
{
   GtkStyle*	pStyle;

  /*
   * Change the color of the widget indirectly
   * by changing the color in the style struct.
   */
   pStyle = gtk_widget_get_style( pWidget );
   pStyle = gtk_style_copy( pStyle );

   pStyle->fg[GTK_STATE_SELECTED] = *pColorFg;
   pStyle->bg[GTK_STATE_SELECTED] = *pColorBg;

   gtk_widget_set_style( pWidget, pStyle );

   return;
}

/******************************************************************************
 * Change the font of a widget.
 *
 * Parms   : pWidget
 *
 *         : pFontName
 *           If "" or NULL then does nothing.
 *
 * Returns : True if successful.
 ******************************************************************************/
gboolean
ChangeStyleFont( GtkWidget* pWidget,
                 char*      pFontName )
{
   GtkStyle*	pStyle;
   GdkFont*		pFont;

  /*
   * Abort if the font name (or widget) is nothing.
   */
   if ( pFontName == NULL  ||  ! pFontName[0]  ||  pWidget == NULL ) return FALSE;

  /*
   * Change the font style.
   * Just return if no font matches the specified font name.
   */
   pStyle = gtk_style_copy( gtk_widget_get_style( pWidget ) );

   pFont = gdk_font_load( pFontName );
   if ( ! pFont ) return FALSE;

   pStyle->font = pFont;
   gtk_widget_set_style( pWidget,
                         pStyle );

   return TRUE;
}

/*****************************************************************************
 * Apply the fonts in all existing list widgets.
 *
 * Parms	: fontNameList (GLOBAL)
 *            Font name applied to all list widgets.
 *****************************************************************************/
void
ApplyFontListWidgets( void )
{
   if ( fontNameList == NULL  ||  fontNameList[0] == '\0' )
      return;

   ChangeStyleFont( pWidgetClistTags,    fontNameList );
   ChangeStyleFont( pWidgetClistModules, fontNameList );
   ChangeStyleFont( pWidgetClistHistory, fontNameList );
   if ( pWidgetClistFound ) /* absent if no search done yet */
      ChangeStyleFont( pWidgetClistFound,   fontNameList );
}

/*****************************************************************************
 * Get the average height (in pixels) of a font.
 *****************************************************************************/
gint
FontHeight( GdkFont* pFont )
{
   static const char str[] = "(|/\\!AabcZzqgj,.";

   g_assert( pFont );

   return gdk_text_height( pFont,
                           str,
                           sizeof(str) );
}

/*****************************************************************************
 * Position a widget at middle of main widget.
 *
 * Parms   : pWidget
 *****************************************************************************/
void
PositionWidgetAtMiddle( GtkWidget* pWidget )
{
   gint		x_pos  = 100;		/* default X */
   gint		y_pos  = 100;		/* default Y */

   gdk_window_get_position( pWidgetMain->window,
                            &x_pos,
                            &y_pos );
   gtk_widget_set_uposition( GTK_WIDGET(pWidget),
                             x_pos + DIALOG_X_OFFSET,
                             y_pos + DIALOG_Y_OFFSET );

   return;
}

/*****************************************************************************
 * Position/show dialog at middle of main widget.
 *
 * Parms   : pWidgetDialog
 *****************************************************************************/
void
PositionShowDialogAtMiddle( GtkWidget* pWidgetDialog )
{
   PositionWidgetAtMiddle( pWidgetDialog );
   gtk_widget_show_all( pWidgetDialog );

   return;
}

/*****************************************************************************
 * Specifically for find dialogs, position/show find dialog widget.
 *
 * Depending on the specified positioning method, restores the previous
 * position of the dialog widget (as recorded when widget was closed).
 *
 * Parms   : pWidgetDialog
 *
 *         : howToPosition
 *           If true, then the dialog is simply positioned roughly at
 *           the middle of the main widget, which is appropriate for popups.
 *           If false, then the dialog is positioned by this function,
 *           in a sophisticated way, which is appropriate for find dialogs.
 *
 * Notes   : Workaround for a GTK quirk:
 *           Calling gtk_widget_set_uposition() before a widget is shown
 *           results in the widget being at a Y co-ord that is slightly
 *           greater than what was intended.  A workaround is to again call
 *           gtk_widget_set_uposition() after the widget is shown.
 *****************************************************************************/
void
PositionShowFindDialog( GtkWidget* pWidgetDialog,
                        gboolean   howToPosition )
{
   gint		x_pos  = 100;		/* default X */
   gint		y_pos  = 100;		/* default Y */

  /*
   * Record how the dialog wil be positioned.
   */
   positionDialogAtMiddle = howToPosition;

  /*
   * Position the dialog at the middle?
   */
   if ( positionDialogAtMiddle )
   {
      PositionShowDialogAtMiddle( pWidgetDialog );
      return;
   }

  /*
   * Either compute an initial position, or restore the previous position.
   */
   if ( findGeom[0] == 0  &&  findGeom[1] == 0 )
   {
     /*
      * Initially position the dialog so that it doesn't obscure the text widget.
      */
      gdk_window_get_position( pWidgetMain->window,
                               &x_pos,
                               &y_pos );
      if ( ! textLeft )
      {
        /*
         * Position the dialog widget flush at the top-left corner.
         */
         gtk_widget_set_uposition( GTK_WIDGET(pWidgetDialog),
                                   x_pos,
                                   y_pos );
        /*
         * Workaround GTK quirk (see above Notes).
         */
         gtk_widget_show_all( pWidgetDialog );
         gtk_widget_set_uposition( GTK_WIDGET(pWidgetDialog),
                                   x_pos,
                                   y_pos );
        /*
         * KLUDGE: Try to configure WidgetGeom() according to window manager.
         */
         KludgeForWidgetGeom( pWidgetDialog,
                              x_pos,
                              y_pos );

         return;
      }
      else
      {
        /*
         * Position the dialog widget flush at the top-right corner.
         */
         int	wMain, hMain, wDialog, hDialog;

        /*
         * In order to successfully get the size of the dialog widget,
         * it must be showing, but a dilemma is that it hasn't been positioned yet.
         * Just position the widget off-screen (to avoid widget flicker).
         */
         gtk_widget_set_uposition( GTK_WIDGET(pWidgetDialog),
                                   5000,
                                   5000 );
         gtk_widget_show_all( pWidgetDialog );

        /*
         * Really position the dialog.
         */
         gdk_window_get_size( pWidgetMain->window,
                              &wMain,
                              &hMain );
         gdk_window_get_size( pWidgetDialog->window,
                              &wDialog,
                              &hDialog );
         gtk_widget_set_uposition( GTK_WIDGET(pWidgetDialog),
                                   x_pos + wMain - wDialog,
                                   y_pos );
        /*
         * KLUDGE: Try to configure WidgetGeom() according to window manager.
         */
         KludgeForWidgetGeom( pWidgetDialog,
                              x_pos + wMain - wDialog,
                              y_pos );

         return;
      }
   }
   else
   {
     /*
      * Restore the previous geom of the dialog.
      */
      gtk_widget_set_uposition( GTK_WIDGET(pWidgetDialogFind),
                                findGeom[0],
                                findGeom[1] );
#if 0
/* Skip restoring size because size of dialog varies when searching for text, tags, etc. */
      gtk_widget_set_usize( GTK_WIDGET(pWidgetDialogFind),
                            findGeom[2],
                            findGeom[3] );
#endif
     /*
      * Workaround GTK quirk (see above Notes).
      */
      gtk_widget_show_all( pWidgetDialogFind );
      gtk_widget_set_uposition( GTK_WIDGET(pWidgetDialogFind),
                                findGeom[0],
                                findGeom[1] );
     return;
   }
}

/*****************************************************************************
 * Return the width/height of the window of a widget.
 * (Evidently there is no GTK function to do so?)
 *
 * Parms   : pWidget
 *
 *         : pWidth (OUT)
 *
 *         : pHeight (OUT)
 *****************************************************************************/
void
WidgetWidthHeight( GtkWidget* pWidget,
                   gint*      pWidth,
                   gint*      pHeight )
{
   gint		x;
   gint		y;
   gint		depth;

   gdk_window_get_geometry( pWidget->window,
                            &x,
                            &y,
                            pWidth,
                            pHeight,
                            &depth );

   return;
}

/*****************************************************************************
 * KLUDGE:
 * So that widget positions can be properly saved/restored, this function
 * tries to indirectly determine the effect of border decorations by
 * comparing (x,y) positions of a parent and child widget.
 * The result is a flag to tell WidgetGeom() which position function to call.
 * This kludge is necessary because GTK+ 1.2.8 lacks a single reliable
 * means to set/get the position of a widget independently of window manager.
 *
 * Parms   : pWidget
 *           (Pass pWidgetFindDialog).
 *
 *           x, y
 *           Pass the desired position of widget (pWidget).
 *
 *           kludgeWidgetGeomUseGetPosition (OUT/GLOBAL)
 *           TRUE  if gdk_window_get_position()    should be called by WidgetGeom().
 *           FALSE if gdk_window_get_root_origin() should be called by WidgetGeom().
 *****************************************************************************/
void
KludgeForWidgetGeom( GtkWidget* pWidget,
                     int x,
                     int y )
{
   int x_dial;
   int y_dial;
   int x_main;
   int y_main;

   g_assert( pWidget && pWidget->window );

   ProcessPendingEvents();
   gdk_window_get_root_origin( pWidget->window,
                               &x_dial,
                               &y_dial );
   gdk_window_get_root_origin( pWidgetMain->window,
                               &x_main,
                               &y_main );
   ProcessPendingEvents();

   kludgeWidgetGeomUseGetPosition = ( y_dial == y_main );

#if 0
#ifdef HYPERSRC_DEBUG
   printf( "desired  x = %d y = %d \n", x, y );
   printf( "dial     x = %d y = %d \n", x_dial, y_dial );
   printf( "main     x = %d y = %d \n", x_main, y_main );
#endif
#endif

   return;
}

/*****************************************************************************
 * Return the geom (x:y:w:h) of the window of a widget.
 * (Evidently there is no GTK function to do so?)
 *
 * Parms   : pWidget
 *
 *         : pX (OUT)
 *
 *         : pY (OUT)
 *
 *         : pW (OUT)
 *           width
 *
 *         : pH (OUT)
 *           height
 *****************************************************************************/
void
WidgetGeom( GtkWidget* pWidget,
            gint*      pX,
            gint*      pY,
            gint*      pW,
            gint*      pH )
{
  /*
   * Compute x:y
   */
   if ( kludgeWidgetGeomUseGetPosition )
   {
     /*
      * This is correct on icewm, incorrect on KDE.
      */
      gdk_window_get_position( pWidget->window,
                               pX,
                               pY );
   }
   else
   {
     /*
      * This is correct on KDE, incorrect on icewm.
      */
      gdk_window_get_root_origin( pWidget->window,
                                  pX,
                                  pY );
   }

  /*
   * Compute w:h
   */
   WidgetWidthHeight( pWidget,
                      pW,
                      pH );

   return;
}

/*****************************************************************************
 * Prepend the string the user typed in a combo widget into a glist.
 * The glist is used by the combo's drop-down widget.
 *
 * Parms   : pString
 *           String which user entered (a copy is used here).
 *
 *           ppGlist (IN/OUT)
 *           Pointer to glist for combo widget drop-down.
 *****************************************************************************/
void
PrependComboDropDown( char*   pStr,
                      GList** ppGList )
{
  /*
   * Only prepend if this string is not same as previous,
   * to avoid populating drop-down with same string when
   * user was repeatedly clicking the Find button.
   *
   * The previous is the first element in the list,
   * but the list might be empty.
   */
   if (    (*ppGList != NULL)
        && (strcmp( pStr, (*ppGList)->data ) == 0) )
      return;

  /*
   * Store a persistent clone of user's string (not the original) in the glist.
   */
   *ppGList = g_list_prepend( *ppGList,
                              CloneString( pStr ) );

  /*
   * Remove the last (oldest) element if the list has become too long.
   */
   if ( g_list_length( *ppGList ) > MAX_DROP_DOWN_STRINGS )
   {
     /*
      * Note that g_list_remove() won't always work as expected,
      * because if the last string is duplicated, then non-last
      * duplicate will be removed because it is before the last.
      */
      *ppGList = \
      g_list_remove_link( *ppGList,
                          g_list_last( *ppGList ) );
   }

   return;
}

/*****************************************************************************
 * Create a button with a label with underlined chars.
 * (GTK has no such function).
 *
 * Parms   : argc, argv
 *****************************************************************************/
GtkWidget*
CreateButtonLabelUnderlined( const gchar* pLabelText,
                             const gchar* pLabelPattern )
{
   GtkWidget* label_widget;
   GtkWidget* pWidgetButton;

  /*
   * Derived from gtk_button_new_with_label().
   */
  pWidgetButton = gtk_button_new ();
  label_widget = gtk_label_new ( pLabelText );
  gtk_label_set_pattern( GTK_LABEL(label_widget), pLabelPattern );
  gtk_misc_set_alignment (GTK_MISC (label_widget), 0.5, 0.5);
  gtk_container_add (GTK_CONTAINER (pWidgetButton), label_widget);
  gtk_widget_show (label_widget);

  return pWidgetButton;
}

/*****************************************************************************
 * Return true if any of the major outer-level functions are busy.
 *****************************************************************************/
gboolean
IsAnyMajorFunctionBusy( void )
{
   gboolean res;

   res =     busy_FindStringGlobal
          || busy_FindDialogClickedFind
          || busy_PopulateTagsClistWidget
          || busy_GotoTagUnderCursor
          || busy_SearchClistWidget_tag
          || busy_SearchClistWidget_module;

   if ( res ) PrintStatusbar( "(busy)" );

   return res;
}

/*****************************************************************************
 * For while busy, set sensitivity of menu items and toolbar buttons.
 *
 * For consistency/simplicity, this should be called once at the outer level
 * of a major function, rather than being nested.
 *****************************************************************************/
void
SetSensitivityWhileBusy( gboolean sensitive )
{
   static int cnt = 0;

  /*
   * In case this instance is nested, do nothing.
   */
   if ( ! sensitive )
   {
      ++cnt;
      g_assert( cnt >= 0 );
      if ( cnt != 1 ) return;
   }
   else
   {
      --cnt;
      g_assert( cnt >= 0 );
      if ( cnt != 0 ) return;
   }

  /*
   * Menu items.
   */
#if 0
   /* --this won't work-- */
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Sort" ), sensitive );
#else
g_assert( pItemFactory );
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Sort/Tags By Name" ), sensitive );
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Sort/Tags By Module" ), sensitive );
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Sort/Tags By Type" ), sensitive );

   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Search/Find Text  F3" ), sensitive );
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Search/Find Tag   F4" ), sensitive );
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Search/Find Module" ), sensitive );

   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Prefs/Edit Prefs" ), sensitive );
   gtk_widget_set_sensitive( gtk_item_factory_get_widget( pItemFactory, "/Prefs/Save Prefs" ), sensitive );
#endif

  /*
   * Toolbar buttons.
   */
g_assert( pWidgetToolbar );
   gtk_widget_set_sensitive( GTK_WIDGET(pWidgetToolbarButtonGotoLine), sensitive );
   gtk_widget_set_sensitive( GTK_WIDGET(pWidgetToolbarButtonGotoTag), sensitive );
   gtk_widget_set_sensitive( GTK_WIDGET(pWidgetToolbarButtonEarlierTag), sensitive );
   gtk_widget_set_sensitive( GTK_WIDGET(pWidgetToolbarButtonLaterTag), sensitive );
   gtk_widget_set_sensitive( GTK_WIDGET(pWidgetToolbarButtonRecordCursor), sensitive );

   return;
}

/*****************************************************************************
 * Set the initial position of the paned slider.
 *****************************************************************************/
void
SetPanedSliderInitialPosition( void )
{
   int		w, h;

  /*
   * If the position wasn't saved in the GNOME config,
   * then compute a default position.
   * (A wrong w/h will be returned if windows are not yet shown.)
   */
   WidgetWidthHeight( pWidgetMain, &w, &h );

   if ( panedSliderPercent == 0 )
      panedSliderPercent = textLeft ? 65 : 30;

   gtk_paned_set_position( GTK_PANED(pWidgetPaned), w * panedSliderPercent / 100 );
}

/*****************************************************************************
 * Return the percentage of the movement of the paned slider that
 * divides the list and text widgets.
 * 0% : left
 * Returns an integer: {0..99}.
 *****************************************************************************/
int
PanedSliderPercent( void )
{
   int		geom[4];
   int		geom2[4];
   float	f1, f2;
   int		percent;

   WidgetGeom( pWidgetNotebook,
               &geom[0], &geom[1], &geom[2], &geom[3] );
   WidgetGeom( pWidgetText,
               &geom2[0], &geom2[1], &geom2[2], &geom2[3] );
   if ( ! textLeft )
   {
      f1 = geom[2];
      f2 = geom2[2];
   }
   else
   {
      f1 = geom2[2];
      f2 = geom[2];
   }
   percent = (f1 / (f1 + f2)) * 100.0;
   g_assert( percent >= 0  &&  percent <= 100 );
   return percent;
}
